/* Copyright (C) 1995 John C. Bowman (bowman@hagar.ph.utexas.edu).
Inline floating-point GCC routines for the 80387/80486 (Version 1.5).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Notes:

1.	The accuracy of the sinh, tanh, asinh, and atanh functions have been
	dramatically improved (relative to the poorly written libm versions)
	for arguments of small magnitude. For example, sinh(1.0e-7) now returns
	the correct value 1.0000000000000017e-07 (the libm routine incorrectly
	returns 9.9999999999975326e-08). The error in the libm routine can
	become very large: it returns 9.7578195523695399e-19 for sinh(1.0e-18)
	(the correct value is 1.0e-18).

2.	The inverse hyperbolic functions (and other functions not strictly part
	of ANSI C) are defined in extensions.h.	The functions coshm1(x)=cosh(x)-1
	and acosh1p(x)=acosh(1+x) are exact inverses of each other (to within
	machine precision), unlike acosh(x) and cosh(x). Inlined versions of
	sgn(x)=(x == 0 ? 0 : (x > 0 ? 1 : -1)) and sgn1(x)=(x >= 0 ? 1 : -1)
	and many other libm routines are also defined in extensions.h.

3.	Inlining may be disabled at compile time with the GCC option
	-D__NO_MATH_INLINES.

4.	Extensions to ANSI C may be disabled with the GCC options -ansi or
	-D__NO_MATH_EXTENSIONS.
*/

#ifndef __MATH_H
#define __MATH_H

#ifdef	__GNUC__
#ifndef __NO_MATH_INLINES

#ifdef __cplusplus
#define	__MATH_INLINE __inline
#else
#define	__MATH_INLINE extern __inline
#endif

__MATH_INLINE double cos(const double);
__MATH_INLINE double sin(const double);

__MATH_INLINE double __expm1(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fldl2e				# e^x - 1 = 2^(x * log2(e)) - 1\n\t"
	"fmul	%%st(1)			# x * log2(e)\n\t"
	"fstl	%%st(1)\n\t"
	"frndint				# int(x * log2(e))\n\t"
	"fxch\n\t"
	"fsub	%%st(1)			# fract(x * log2(e))\n\t"
	"f2xm1				# 2^(fract(x * log2(e))) - 1\n\t"
	"fscale				# 2^(x * log2(e)) - 2^(int(x * log2(e)))\n\t"
	"fxch\n\t"
	"fld1\n\t"
	"fscale				# 2^int(x * log2(e))\n\t"
	"fstpl	%%st(1)\n\t"
	"fld1\n\t"
	"fsubrp				# 2^int(x * log2(e)) - 1\n\t"
	"faddp				# e^x - 1"
	: "=t" (__value) : "0" (__arg));

	return __value;
}

__MATH_INLINE double __sgn1(const double x)
{
	return (x >= 0.0 ? 1.0 : -1.0);
}

__MATH_INLINE double sqrt(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fsqrt" : "=t" (__value): "0" (__arg));

	return __value;
}

__MATH_INLINE double fabs(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fabs" : "=t" (__value): "0" (__arg));

	return __value;
}

__MATH_INLINE double sin(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fsin" : "=t" (__value): "0" (__arg));

	return __value;
}

__MATH_INLINE double cos(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fcos" : "=t" (__value): "0" (__arg));

	return __value;
}

__MATH_INLINE double tan(const double x)
{
	register double __value, __value2, __arg = x;
	__asm__ __volatile__
	  ("fptan" : "=t" (__value2), "=u" (__value): "0" (__arg));

	return __value;
}

__MATH_INLINE double atan2(const double y, const double x)
{
	register double __value, __argy = y, __argx = x;
	__asm__ __volatile__
	  ("fld %%st(1)\n\t"
	   "fpatan"
	: "=t" (__value): "0" (__argy), "u" (__argx));

	return __value;
}


__MATH_INLINE double asin(const double x)
{
	register double __asin = x;

	return atan2(__asin,sqrt(1.0-__asin*__asin));
}

__MATH_INLINE double acos(const double x)
{
	register double __acos = x;

	return atan2(sqrt(1.0-__acos*__acos),__acos);
}

__MATH_INLINE double atan(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fld1\n\t"
	   "fpatan"
	: "=t" (__value): "0" (__arg));

	return __value;
}

__MATH_INLINE double exp(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fldl2e				# e^x = 2^(x * log2(e))\n\t"
	"fmul	%%st(1)			# x * log2(e)\n\t"
	"fstl	%%st(1)\n\t"
	"frndint				# int(x * log2(e))\n\t"
	"fxch\n\t"
	"fsub	%%st(1)			# fract(x * log2(e))\n\t"
	"f2xm1				# 2^(fract(x * log2(e))) - 1\n\t"
	"fld1\n\t"
	"faddp				# 2^(fract(x * log2(e)))\n\t"
	"fscale				\n\t"
	"fstpl	%%st(1)			# e^x"
	: "=t" (__value) : "0" (__arg));

	return __value;
}

__MATH_INLINE double sinh(const double x)
{
	register double __sinharg = x, __sinh = __expm1(fabs(__sinharg));

	return 0.5*(__sinh/(__sinh+1.0)+__sinh)*__sgn1(__sinharg);
}

__MATH_INLINE double cosh(const double x)
{
	register double __cosh = exp(x);

	return 0.5*(__cosh+1.0/__cosh);
}

__MATH_INLINE double tanh(const double x)
{
	register double __tanharg = x, __tanh = __expm1(-fabs(2.0*__tanharg));

	return __tanh/(__tanh+2.0)*__sgn1(-__tanharg);
}

__MATH_INLINE double log(const double x)
{
	register double __value, __arg = x ;
	__asm__ __volatile__
	  ("fldln2\n\t"
	   "fxch\n\t"
	   "fyl2x"
	: "=t" (__value) : "0" (__arg));

	return __value;
}

__MATH_INLINE double log10(const double x)
{
	register double __value, __arg = x ;
	__asm__ __volatile__
	  ("fldlg2\n\t"
	   "fxch\n\t"
	   "fyl2x"
		: "=t" (__value) : "0" (__arg));

	return __value;
}

__MATH_INLINE double fmod(const double x, const double y)
{
	register double __value, __value2, __argx = x, __argy = y;
	__asm__ __volatile__
	  ("1:	fprem\n\t"
	"fstsw	%%ax\n\t"
	"sahf\n\t"
	"jp	1b"
	: "=t" (__value), "=u" (__value2)
	: "0" (__argx), "1" (__argy) : "ax", "cc");

	return __value;
}

__MATH_INLINE double ldexp(const double x, const int y)
{
	register double __value, __argx = x, __argy = (double)y;
	__asm__ __volatile__
	  ("fscale" : "=t" (__value): "0" (__argx), "u" (__argy));

	return __value;
}

#ifndef __NO_MATH_EXTENSIONS
#include <i386/extensions.h>
#endif

#endif /* __NO_MATH_INLINES */
#endif	/* __GNUC__	 */

#endif /* __MATH_H */
