/* Copyright (C) 1995 John C. Bowman (bowman@hagar.ph.utexas.edu).
Extensions to inline floating-point routines for the 80387/80486 (Version 1.5).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

__MATH_INLINE double __hypot(const double x, const double y)
{
	register double __argx = x, __argy = y;

	return sqrt(__argx*__argx+__argy*__argy);
}

__MATH_INLINE double __log1p(const double x)
{
	register double __value, __arg = x;

	if (fabs (__arg) >= 1.0-0.5*M_SQRT2)
		__value = log(1.0+__arg);

	else
		__asm__ __volatile__
		  ("fldln2\n\t"
		  "fxch\n\t"
		  "fyl2xp1"
		: "=t" (__value) : "0" (__arg));

	return __value;
}

__MATH_INLINE double __asinh(const double x)
{
	register double __asinharg = x, __asinh=fabs(__asinharg),
					   __asinhI=1.0/__asinh;

	return __log1p(__asinh/(sqrt(__asinhI*__asinhI+1.0)+__asinhI)+__asinh)
			*__sgn1(__asinharg);
}

__MATH_INLINE double __acosh(const double x)
{
	register double __acosh = x;

	return log(__acosh+sqrt(__acosh-1.0)*sqrt(__acosh+1.0));
}

__MATH_INLINE double __atanh(const double x)
{
	register double __atanharg = x, __atanh=fabs(__atanharg);

	return -0.5*__log1p(-2.0*__atanh/(1.0+__atanh))*__sgn1(__atanharg);
}

__MATH_INLINE double __coshm1(const double x)
{
	register double __cosh = __expm1(fabs(x));

	return 0.5*(__cosh/(__cosh+1.0))*__cosh;
}

__MATH_INLINE double __acosh1p(const double x)
{
	register double __acosh = x;

	return __log1p(__acosh+sqrt(__acosh)*sqrt(__acosh+2.0));
}

__MATH_INLINE double __logb(const double x)
{
	register double __value, __arg = x;
	__asm__ __volatile__
	  ("fxtract\n\t"
	   "fstpl	%%st"
	: "=t" (__value): "0" (__arg));

	return	__value;
}

__MATH_INLINE double __drem(const double x, const double y)
{
	register double __value, __value2, __argx = x, __argy = y;
	__asm__ __volatile__
	  ("1:	fprem1\n\t"
	"fstsw	%%ax\n\t"
	"sahf\n\t"
	"jp	1b" : "=t" (__value), "=u" (__value2)
	: "0" (__argx), "1" (__argy) : "ax", "cc");

	return __value;
}

__MATH_INLINE void __sincos(const double x, double *sinx, double *cosx)
{
	register double __cos, __sin, __arg = x;
	__asm__ __volatile__
	  ("fsincos": "=t" (__cos), "=u" (__sin) : "0" (__arg));

	*sinx = __sin; *cosx = __cos;
}

__MATH_INLINE double __sgn(const double x)
{
	return (x == 0.0 ? 0.0 : (x > 0.0 ? 1.0 : -1.0));
}

#ifndef __STRICT_ANSI__

#if __OPTIMIZE__

#define expm1 __expm1
#define hypot __hypot
#define log1p __log1p
#define asinh __asinh
#define acosh __acosh
#define atanh __atanh
#define coshm1 __coshm1
#define acosh1p __acosh1p
#define logb __logb
#define drem __drem
#define sincos __sincos
#define sgn __sgn
#define sgn1 __sgn1

#else

#define sincos(x,sinx,cosx) (*sinx=sin(x), *cosx=cos(x))
#define sgn(x) ((x) == 0.0 ? 0.0 : ((x) > 0.0 ? 1.0 : -1.0))
#define sgn1(x) ((x) >= 0.0 ? 1.0 : -1.0)

#endif /* __OPTIMIZE__ */

#endif /* __STRICT_ANSI__ */
